package nvc

import (
	"context"
	"fmt"
	"strings"

	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"

	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

// Generate the name of migrate-spare deployment for VM right sizing
func GetQuarkMigrationSpareDeployName(volumeUUID string) string {
	return fmt.Sprintf("%s-%s", quarkconfig.MigrationSpare, volumeUUID)
}

// To migrate quark pod to a right-sized VM:
// 1. create a migration-spare deployment
// 2. returns nil if quark deployment has been updated with new node selctor, meaning migration of quark pod has been initiated
func (c *Controller) migrateQuarkPodVMRightSizing(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	nodeSelector map[string]string,
) error {
	var (
		namespace, name = GetNamespaceNameForNetappObj(volume, storagePool)
	)

	// when we are here, there's a possibility that deployment is updated but pod is yet to be created on the new node
	// migration isn't needed at that point and we simply return and wait for pod to be migrated to new node
	updateDeployment, err := c.needDeploymentUpdate(namespace, name, nodeSelector)
	if err != nil {
		return err
	}
	// check VM right sizing enabled and create migration-spare deployment when node selectors do not match
	if serviceLevel.Spec.RightSize && updateDeployment {
		err := c.deployMigrationSpare(nodeSelector, volume, storagePool, serviceLevel)
		if err != nil {
			return err
		}

		// update the quark deployment only after migration-spare pod is up and running
		// Fetch the latest deployment from Kuberenetes and update the nodeselector
		quarkDeployment, err := c.KubeClientset.AppsV1().Deployments(namespace).Get(context.TODO(), GetQuarkDeployName(namespace, name), metav1.GetOptions{})
		if err != nil {
			return fmt.Errorf("failed to get quark deployment for VM right sizing: %s", err.Error())
		}
		quarkDeployment.Spec.Template.Spec.NodeSelector = nodeSelector

		// Get bucket from size and service level
		quarkPodBucket, err := getServiceLevelBucket(storagePool.Spec.Size.Value(), serviceLevel)
		if err != nil || quarkPodBucket == nil {
			return fmt.Errorf("failed to get quark bucket %v", err)
		}

		dmsCPUs := quarkPodBucket.DmsCPUs.Value()

		// Get container resources map from the quark pod bucket
		containerResourcesMap := util.GetContainerResourcesMap(quarkPodBucket)

		// Update container resources in deployment
		for i := range quarkDeployment.Spec.Template.Spec.Containers {
			if _, ok := containerResourcesMap[quarkDeployment.Spec.Template.Spec.Containers[i].Name]; ok {
				quarkDeployment.Spec.Template.Spec.Containers[i].Resources = corev1.ResourceRequirements{
					Requests: corev1.ResourceList{
						corev1.ResourceCPU:    containerResourcesMap[quarkDeployment.Spec.Template.Spec.Containers[i].Name].CPUsRequest,
						corev1.ResourceMemory: containerResourcesMap[quarkDeployment.Spec.Template.Spec.Containers[i].Name].MemRequest,
					},
					Limits: corev1.ResourceList{
						corev1.ResourceCPU:    containerResourcesMap[quarkDeployment.Spec.Template.Spec.Containers[i].Name].CPUsLimit,
						corev1.ResourceMemory: containerResourcesMap[quarkDeployment.Spec.Template.Spec.Containers[i].Name].MemLimit,
					},
				}

				// For dmap container update DMSCPUs env as well
				if quarkDeployment.Spec.Template.Spec.Containers[i].Name == quarkconfig.QuarkContainerDMAP {
					for j := range quarkDeployment.Spec.Template.Spec.Containers[i].Env {
						if quarkDeployment.Spec.Template.Spec.Containers[i].Env[j].Name == "DMS_CPUS_ENV" {
							quarkDeployment.Spec.Template.Spec.Containers[i].Env[j].Value = fmt.Sprintf("%d", dmsCPUs)
						}
					}
				}
			}
		}

		_, err = c.KubeClientset.AppsV1().Deployments(namespace).Update(context.TODO(),
			quarkDeployment, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update the quark deployment: %s", err.Error())
		}
		log.WithFields(log.Fields{
			"netappObj":    name,
			"nodeSelector": nodeSelector,
		}).Info("Updated the Quark deployment with new node selector and resource requests/limits.")
	}
	return nil
}

func getZoneSelector(
	nodeSelectors map[string]string,
) (key, value string, ok bool) {
	ret, ok := nodeSelectors[quarkconfig.ZoneTopologyKey]
	return quarkconfig.ZoneTopologyKey, ret, ok
}

// Returns true if the expected node selector obtained from service level is not present on the deployment
func (c *Controller) needDeploymentUpdate(namespace string, name string, expectedNodeSelectors map[string]string) (bool, error) {
	quarkDeploy, err := c.deploysLister.Deployments(namespace).Get(GetQuarkDeployName(namespace, name))
	if errors.IsNotFound(err) {
		return false, nil
	} else if err != nil {
		return false, fmt.Errorf("failed to get the quark deployment: %s", err.Error())
	}
	// If the deployment doesn't have at least one of the expected node selectors,
	// we need to update it
	for k, v := range expectedNodeSelectors {
		if v != quarkDeploy.Spec.Template.Spec.NodeSelector[k] {
			log.WithFields(log.Fields{
				"name":                  name,
				"expectedNodeSelectors": expectedNodeSelectors,
				"nodeSelectors":         quarkDeploy.Spec.Template.Spec.NodeSelector,
			}).Debug("Found node selector mismatch.")
			return true, nil
		}
	}
	return false, nil
}

func (c *Controller) needsResizeMigration(
	pod *corev1.Pod, expectedNodeSelectors map[string]string,
) (bool, error) {
	// Check if the pod already has the node selector
	labelKey, expectedValue, ok := util.GetVMRightSizingSelector(expectedNodeSelectors)
	if !ok {
		// No VMRS node selector
		return false, nil
	}
	currentValue, ok := pod.Spec.NodeSelector[labelKey]
	if ok {
		return (currentValue != expectedValue), nil
	}

	// Pod doesn't have the node selector, check if the node has it
	nodeName := pod.Spec.NodeName
	if len(nodeName) == 0 {
		// if pod doesn't have a node name yet, it's getting one assigned, skip migration
		return false, nil
	}
	currentValue, err := c.getNodeLabelValue(nodeName, labelKey)
	if err != nil {
		return false, err
	}
	return (currentValue != expectedValue), nil
}

func (c *Controller) needsZoneMigration(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	pod *corev1.Pod,
	expectedNodeSelectors map[string]string,
) (bool, error) {
	if storagePool.Spec.LastZoneMigrationCause != quarkconfig.ZoneMigrationCauseUser {
		// Don't do a graceful migration if the sp was migrated because of a zone
		// failure; the NHC is on it.
		return false, nil
	}
	labelKey, expectedValue, ok := getZoneSelector(expectedNodeSelectors)
	if !ok {
		return false, nil
	}
	currentValue, ok := pod.Spec.NodeSelector[labelKey]
	if ok {
		return (currentValue != expectedValue), nil
	}

	// Pod doesn't have the node selector, check if the node has it
	nodeName := pod.Spec.NodeName
	if nodeName == "" {
		// if pod doesn't have a node name yet, it's getting one assigned, skip migration
		return false, nil
	}
	currentValue, err := c.getNodeLabelValue(nodeName, labelKey)
	if err != nil {
		return false, err
	}
	return currentValue != expectedValue, nil
}

// Returns true if pod needs to be migrated to the right sized VM during volume
// resize or upgrade, or if it needs to be migrated across zones due to user
// action.
func (c *Controller) needQuarkPodMigration(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, expectedNodeSelectors map[string]string) (bool, error) {
	var needsZoneMigration = false
	if c.QuarkVariant == "min" {
		// In the min config there is only one node, so there is no migration
		return false, nil
	}
	if c.regionalMeta != nil && storagePool != nil && storagePool.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseUser && storagePool.Status.Zone != "" && storagePool.Spec.Zone != storagePool.Status.Zone {
		cm, err := c.KubeClientset.CoreV1().ConfigMaps(storagePool.Namespace).Get(
			context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
		if err != nil {
			return false, fmt.Errorf("unable to retrieve regional HA configmavp: %v",
				err)
		}
		if cm.Data[storagePool.Spec.Zone] == quarkconfig.RegionalHAActiveZone {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"targetZone":  storagePool.Spec.Zone,
				"currentZone": storagePool.Status.Zone,
			}).Debug("Migrating SP due to zone mismatch.")
			// If there's a mismatch between the spec and status zone, we need to
			// migrate zones, regardless of whether a resize is needed, so return true
			// immediately. Don't do this in the event of a zone failure; it's already
			// covered by the NHC.
			return true, nil
		} else {
			// TODO: Does this end up being dangerous when there's simultaneous resize
			// and migration into a failed zone?
			delete(expectedNodeSelectors, quarkconfig.ZoneTopologyKey)
		}
	}
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	podList, err := c.podsLister.Pods(namespace).List(GetQuarkPodSelector(volume, storagePool))
	if err != nil {
		return false, err
	}
	if len(podList) == 0 {
		// No pods, check if deployment needs update
		return c.needDeploymentUpdate(namespace, name, expectedNodeSelectors)
	}

	// There's only one quark pod per volume
	pod := podList[0]

	needsResizeMigration, err := c.needsResizeMigration(
		pod, expectedNodeSelectors)
	if err != nil {
		return false, err
	}
	if storagePool != nil && c.regionalMeta != nil {
		// Only migrate zones if we have a storage pool
		needsZoneMigration, err = c.needsZoneMigration(storagePool, pod,
			expectedNodeSelectors)
		if err != nil {
			return false, err
		}
	}
	if needsResizeMigration {
		var name, objType = "", "volume"
		if storagePool != nil {
			name = storagePool.Name
			objType = "storagePool"
		} else {
			name = volume.Name
		}
		log.WithFields(log.Fields{
			objType: name,
		}).Debug("Needs resize migration.")
	}
	if needsZoneMigration {
		//storagePool is guaranteed non nil here
		log.WithFields(log.Fields{
			"storagePool":           storagePool.Name,
			"targetZone":            storagePool.Spec.Zone,
			"currentZone":           storagePool.Status.Zone,
			"expectedNodeSelectors": expectedNodeSelectors,
			"nodeSelectors":         pod.Spec.NodeSelector,
		}).Debug("Migrating SP because node selectors don't match.")
	}
	return (needsResizeMigration || needsZoneMigration), nil
}

func (c *Controller) GetNode(nodeName string) (*corev1.Node, error) {
	return c.nodesLister.Get(nodeName)
}

// Retrieve label value from a node
func (c *Controller) getNodeLabelValue(nodeName, labelKey string) (string, error) {
	node, err := c.GetNode(nodeName)
	if err != nil || node.ObjectMeta.DeletionTimestamp != nil {
		return "", err
	}
	return node.Labels[labelKey], nil
}

// Create migration-spare deployment and return nil only if the pod is ready
func (c *Controller) deployMigrationSpare(nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) error {
	var quarkVersion string
	if storagePool != nil {
		quarkVersion = storagePool.Spec.QuarkVersion
	} else {
		quarkVersion = volume.Spec.QuarkVersion
	}

	imageLookup := c.getImageTags(quarkVersion)
	msdeploy, err := c.reconcileMigrationSpareDeployment(imageLookup, nodeSelector, volume, storagePool, serviceLevel)
	if err != nil {
		return err
	}
	if msdeploy != nil {
		if msdeploy.Status.ReadyReplicas > 0 {
			return nil
		}
	}
	return MigrationInProgressError{}
}

func (c *Controller) reconcileMigrationSpareDeployment(imageMap, nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*appsv1.Deployment, error) {
	var (
		updatedNodeSelector bool
		namespace, name     = GetNamespaceNameForNetappObj(volume, storagePool)
	)
	deploymentName := GetQuarkMigrationSpareDeployName(name)
	msdeploy, err := c.deploysLister.Deployments(namespace).Get(deploymentName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, fmt.Errorf("failed to get deployment %s: %s", deploymentName, err.Error())
	}
	if msdeploy != nil {
		if msdeploy.DeletionTimestamp != nil {
			return nil, fmt.Errorf("failed to reconcile deployment %s: undergoing deletion", deploymentName)
		} else {
			// update the existing migration-spare deployment, if node-selector values present on it do not match with the desired values
			msNodeSelector := msdeploy.Spec.Template.Spec.NodeSelector
			ms := msdeploy.DeepCopy()
			for k, v := range nodeSelector {
				if val, ok := msNodeSelector[k]; ok && val != v {
					ms.Spec.Template.Spec.NodeSelector[k] = nodeSelector[k]
					updatedNodeSelector = true
				}
			}
			if updatedNodeSelector {
				if msdeploy, err = c.KubeClientset.AppsV1().Deployments(namespace).Update(context.TODO(),
					ms, metav1.UpdateOptions{}); err != nil {
					return nil, fmt.Errorf("failed to update deployment %s: %s", deploymentName, err.Error())
				}
			}
		}
	} else {
		ms := c.makeMigrationSpareDeployment(imageMap, nodeSelector, volume, storagePool, serviceLevel)
		msdeploy, err = c.KubeClientset.AppsV1().Deployments(namespace).Create(context.TODO(),
			ms, metav1.CreateOptions{})
		if err != nil && !errors.IsAlreadyExists(err) {
			return nil, fmt.Errorf("failed to create deployment %s: %s", deploymentName, err.Error())
		}
	}
	return msdeploy, nil
}

func (c *Controller) makeMigrationSpareDeployment(imageMap, nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) *appsv1.Deployment {
	var (
		labelSelector metav1.LabelSelector
		labels        map[string]string
		affinityTerm  *corev1.Affinity
	)

	_, name := GetNamespaceNameForNetappObj(volume, storagePool)
	gracePeriodSeconds := int64(0)
	replicas := int32(1)
	resources := getMigrationSpareResources(serviceLevel, nodeSelector)
	isSharedProcess := true
	if storagePool == nil {
		labels = map[string]string{
			quarkconfig.AppLabel:    quarkconfig.MigrationSpare,
			quarkconfig.VolumeLabel: name,
		}
		labelSelector = metav1.LabelSelector{
			MatchLabels: map[string]string{
				quarkconfig.VolumeLabel: name,
			},
		}
	} else {
		labels = map[string]string{
			quarkconfig.AppLabel:         quarkconfig.MigrationSpare,
			quarkconfig.StoragePoolLabel: name,
		}
		labelSelector = metav1.LabelSelector{
			MatchLabels: map[string]string{
				quarkconfig.StoragePoolLabel: name,
			},
		}
	}
	if _, ok := nodeSelector[quarkconfig.ZoneTopologyKey]; !ok {
		// Only add affinity for the Quark pod if we don't already have a defined
		// zone
		affinityTerm = &corev1.Affinity{
			PodAffinity: &corev1.PodAffinity{
				RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
					{
						LabelSelector: &labelSelector,
						TopologyKey:   quarkconfig.ZoneTopologyKey,
					},
				},
			},
		}
	}

	// Add nodevol to the migration spare
	source := corev1.HostPathDirectory
	quarkVolume := corev1.Volume{
		Name: "quarkvol",
		VolumeSource: corev1.VolumeSource{
			HostPath: &corev1.HostPathVolumeSource{
				Type: &source,
				Path: quarkconfig.NodevolHostpath,
			},
		},
	}

	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					NodeSelector:                  nodeSelector,
					ServiceAccountName:            quarkconfig.MigrationSpareServiceAccount,
					PriorityClassName:             quarkconfig.NegativePriorityClass,
					TerminationGracePeriodSeconds: &gracePeriodSeconds,
					ShareProcessNamespace:         &isSharedProcess,
					Affinity:                      affinityTerm,

					Containers: makeContainers(imageMap, resources, c.Cloud),
					Volumes:    []corev1.Volume{quarkVolume},
				},
			},
		},
	}
	deploy.Name = GetQuarkMigrationSpareDeployName(name)
	deploy.Labels = labels
	return deploy
}

func getMigrationSpareResources(serviceLevel *quarkv1alpha1.NetAppServiceLevel, nodeSelector map[string]string) corev1.ResourceList {
	for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {

		// The pod's node selector may contain more keys than the pod bucket,
		// so we can't do equality.
		// TODO: This will break if we ever add additional keys to the pod
		// bucket node selector.
		if strings.Contains(util.NodeSelectorMapToString(nodeSelector),
			qpb.NodeSelector) {
			return util.GetResources(qpb)
		}
	}
	return corev1.ResourceList{}
}

func makeContainers(imageMap map[string]string, resources corev1.ResourceList, cloud string) []corev1.Container {
	containers := util.PrepullContainersFromImageMap(imageMap, cloud)
	// We will have atleast one container in the list always
	if len(containers) > 0 {
		containers[0].VolumeMounts = []corev1.VolumeMount{
			{
				Name:      "quarkvol",
				MountPath: "/mroot",
			},
		}

		containers[0].Resources = corev1.ResourceRequirements{
			Limits:   resources,
			Requests: resources,
		}
	}
	return containers
}

func (c *Controller) reconcileDeleteMigrationSpare(namespace, deploymentName string) error {
	err := c.KubeClientset.AppsV1().Deployments(namespace).Delete(
		context.TODO(), deploymentName, metav1.DeleteOptions{})
	if err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete deployment %s: %s", deploymentName, err.Error())
	} else if err == nil {
		log.WithFields(log.Fields{
			"deploymentName": deploymentName,
		}).Info("Deleting the migration-spare deployment.")
	}
	return nil
}
