package v2

import (
	"fmt"
	"strings"
	"sync"
	"time"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

type VolumeEvent string
type StoragePoolEvent string
type SnapshotEvent string
type CloudSnapshotEvent string
type SnapshotPolicyEvent string
type ExportPolicyEvent string
type QuarkEvent string
type ClusterPeeringEvent string
type VolumeReplicationEvent string

const (
	VolumeOnline                             VolumeEvent         = "VolOnline"
	VolumeOffline                            VolumeEvent         = "VolOffline"
	VolumeResized                            VolumeEvent         = "VolResized"
	VolumeDeleted                            VolumeEvent         = "VolDeleted"
	VolumeVersionUpgraded                    VolumeEvent         = "VolVersionUpgraded"
	VolumeSnapPolicyApplied                  VolumeEvent         = "VolSnapPolApplied"
	VolumeExpPolicyApplied                   VolumeEvent         = "VolExpPolApplied"
	VolumeSMBSharePropertiesApplied          VolumeEvent         = "VolsmbSharePropertiesApplied"
	VolumeSMBPasswordReset                   VolumeEvent         = "VolSMBPasswordReset"
	VolumeSMBDNSUpdated                      VolumeEvent         = "VolSMBDNSUpdated"
	VolumeNoSnapDirOptionUpdated             VolumeEvent         = "VolumeNoSnapDirUpdated"
	StoragePoolSMBPasswordReset              StoragePoolEvent    = "SpSMBPasswordReset"
	StoragePoolSMBDNSUpdated                 StoragePoolEvent    = "SpSMBDNSUpdated"
	StoragePoolSMBADConfigurationUpdated     StoragePoolEvent    = "StoragePoolSMBADConfigurationUpdated"
	StoragePoolSMBADConfigUserListUpdated    StoragePoolEvent    = "StoragePoolSMBADConfigUserListUpdated"
	StoragePoolSMBServerConfigurationUpdated StoragePoolEvent    = "StoragePoolSMBServerConfigurationUpdated"
	StoragePoolLdapConfigured                StoragePoolEvent    = "StoragePoolLdapConfigured"
	StoragePoolOnline                        StoragePoolEvent    = "SpOnline"
	StoragePoolOffline                       StoragePoolEvent    = "SpOffline"
	StoragePoolResized                       StoragePoolEvent    = "SpResized"
	StoragePoolMigratedZone                  StoragePoolEvent    = "SpMigratedZone"
	StoragePoolGlobalAccessUpdated           StoragePoolEvent    = "SpGlobalAccessUpdated"
	StoragePoolDeleted                       StoragePoolEvent    = "SpDeleted"
	StoragePoolVersionUpgraded               StoragePoolEvent    = "SpVersionUpgraded"
	SnapshotReady                            SnapshotEvent       = "SnapReady"
	SnapshotDeleted                          SnapshotEvent       = "SnapDeleted"
	SnapshotRestored                         SnapshotEvent       = "SnapRestored"
	SnapshotsListUpdated                     SnapshotEvent       = "SnapListUpdated"
	CloudSnapshotReady                       CloudSnapshotEvent  = "CloudSnapReady"
	CloudSnapshotDeleted                     CloudSnapshotEvent  = "CloudSnapDeleted"
	SnapshotPolicyUpdated                    SnapshotPolicyEvent = "SnapPolUpdated"
	ExportPolicyUpdated                      ExportPolicyEvent   = "ExpPolUpdated"
	QuarkReady                               QuarkEvent          = "QuarkReady"
	QuarkNotReady                            QuarkEvent          = "QuarkNotReady"
	QuarkOperatorReady                       QuarkEvent          = "QuarkOperatorReady"
	QuarkOperatorNotReady                    QuarkEvent          = "QuarkOperatorNotReady"
	VolumeControllerReady                    QuarkEvent          = "VolumeControllerReady"
	VolumeControllerNotReady                 QuarkEvent          = "VolumeControllerNotReady"
	HighAvailabilityReady                    QuarkEvent          = "HighAvailabilityReady"
	HighAvailabilityNotReady                 QuarkEvent          = "HighAvailabilityNotReady"
	ExternalVolumeAccessReady                QuarkEvent          = "ExternalVolumeAccessReady"
	ExternalVolumeAccessNotReady             QuarkEvent          = "ExternalVolumeAccessNotReady"
	CloudsnapshotCapabilityReady             QuarkEvent          = "CloudsnapshotCapabilityReady"
	CloudsnapshotCapabilityNotReady          QuarkEvent          = "CloudsnapshotCapabilityNotReady"
	IPDReady                                 QuarkEvent          = "IPDReady"
	IPDNotReady                              QuarkEvent          = "IPDNotReady"
	EtcdClusterReady                         QuarkEvent          = "EtcdClusterReady"
	EtcdClusterNotReady                      QuarkEvent          = "EtcdClusterNotReady"
	NodevolReady                             QuarkEvent          = "NodevolReady"
	NodevolNotReady                          QuarkEvent          = "NodevolNotReady"
	QuarkDeleted                             QuarkEvent          = "QuarkDeleted"
	ServiceMeshControllerReady               QuarkEvent          = "ServiceMeshControllerReady"
	ServiceMeshControllerNotReady            QuarkEvent          = "ServiceMeshControllerNotReady"
	EventHandlerQueueSize                                        = 100
	VolumeSMBEncryptedADCommunicationUpdated VolumeEvent         = "VolSMBEncryptedADCommunicationUpdated"
	ClusterPeeringReady                      ClusterPeeringEvent = "ClusterPeeringReady"
	ClusterPeeringDeleted                    ClusterPeeringEvent = "ClusterPeeringDeleted"
	ServiceMeshEnabled                       ClusterPeeringEvent = "ServiceMeshEnabled"
	// Volume Replication events
	VolumeReplicationInitialized        VolumeReplicationEvent = "VolumeReplicationInitialized"
	VolumeReplicationBroken             VolumeReplicationEvent = "VolumeReplicationBroken"
	VolumeReplicationHealthy            VolumeReplicationEvent = "VolumeReplicationHealthy"
	VolumeReplicationDeleted            VolumeReplicationEvent = "VolumeReplicationDeleted"
	VolumeReplicationScheduleUpdated    VolumeReplicationEvent = "VolumeReplicationScheduleUpdated"
	VolumeReplicationStatusIdle         VolumeReplicationEvent = "VolumeReplicationStatusIdle"
	VolumeReplicationStatusTransferring VolumeReplicationEvent = "VolumeReplicationStatusTransferring"
)

type NVCClientTimeoutError struct {
	err string
}

func (te *NVCClientTimeoutError) Error() string {
	return te.err
}

func NewNVCClientTimeoutError(event string) *NVCClientTimeoutError {
	return &NVCClientTimeoutError{fmt.Sprintf("%v event timed out", event)}
}

func NewNVCClientTimeoutErrorStoragePool(eventString string) *NVCClientTimeoutError {
	return &NVCClientTimeoutError{fmt.Sprintf("Internal Error: %v", eventString)}
}

type NVCClientFatalError struct {
	err string
}

func (fe *NVCClientFatalError) Error() string {
	return fe.err
}

func NewNVCClientFatalError(errMessage string) *NVCClientFatalError {
	return &NVCClientFatalError{errMessage}
}

type NetAppVolumeEvent struct {
	Volume *quarkv1alpha1.NetAppVolume
	//TODO: handle multiple events at the same time.
	EventType VolumeEvent
	Channel   chan error
}

type NetAppStoragePoolEvent struct {
	StoragePool *quarkv1alpha1.NetAppStoragePool
	EventType   StoragePoolEvent
	Channel     chan error
}

type NetAppQuarkEvent struct {
	QuarkName string
	EventType QuarkEvent
	Channel   chan error
}

type NetAppVolumeSnapshotEvent struct {
	// VolumeUUID is required
	VolumeUUID string
	//SnapDisplayName is optional
	SnapDisplayName string
	//TODO: handle multiple events at the same time.
	EventType SnapshotEvent
	Channel   chan error
}

type NetAppVolumeCloudSnapshotEvent struct {
	CloudSnapshot string
	//TODO: handle multiple events at the same time.
	EventType CloudSnapshotEvent
	Channel   chan error
}

type NetAppExportPolicyEvent struct {
	ExportPolicy string
	EventType    ExportPolicyEvent
	Channel      chan error
}

type NetAppSnapshotPolicyEvent struct {
	SnapshotPolicy string
	EventType      SnapshotPolicyEvent
	Channel        chan error
}

type NetAppClusterPeeringEvent struct {
	ClusterPeering *quarkv1alpha1.NetAppClusterPeering
	EventType      ClusterPeeringEvent
	Channel        chan error
}

type NetAppVolumeReplicationEvent struct {
	VolumeReplication *quarkv1alpha1.NetAppVolumeReplication
	EventType         VolumeReplicationEvent
	Channel           chan error
}

type NVCClientEventHandler struct {
	// The handle to NVC Client.
	client *NVCClient
	// The channel event handler receives NetAppVolume update events from NVC Client.
	nvcVolUpdateChan <-chan *NetAppVolumeUpdate
	// The channel event handler receives NetAppVolume delete events from NVC Client.
	nvcVolDeleteChan <-chan *NetAppVolumeDelete
	// The channel event handler receives NetAppStoragePool update events from NVC Client.
	nvcSpUpdateChan <-chan *NetAppStoragePoolUpdate
	// The channel event handler receives NetAppStoragePool delete events from NVC Client.
	nvcSpDeleteChan <-chan *NetAppStoragePoolDelete
	// The channel event handler receives NetAppVolumeSnapshot update events from NVC Client.
	nvcSnapUpdateChan <-chan *NetAppVolumeSnapshotUpdate
	// The channel event handler receives NetAppVolumeSnapshot delete events from NVC Client.
	nvcSnapDeleteChan <-chan *NetAppVolumeSnapshotDelete
	// The channel event handler receives NetAppVolumeSnapshot restore events from NVC Client.
	nvcSnapRestoreChan <-chan *NetAppVolumeSnapshotRestore
	// The channel event handler receives NetAppVolumeSnapshotsList update events from NVC Client.
	nvcSnapsListUpdateChan <-chan *NetAppVolumeSnapshotsListUpdate
	// The channel event handler receives NetAppVolumeCloudSnapshot update events from NVC Client.
	nvcCloudSnapUpdateChan <-chan *NetAppVolumeCloudSnapshotUpdate
	// The channel event handler receives NetAppVolumeCloudSnapshot delete events from NVC Client.
	nvcCloudSnapDeleteChan <-chan *NetAppVolumeCloudSnapshotDelete
	// The channel event handler receives NetAppVolumeExportPolicy update events from NVC Client.
	nvcExportPolUpdateChan <-chan *NetAppVolumeExportPolicyUpdate
	// The channel event handler receives NetAppVolumeSnapshotPolicy update events from NVC Client.
	nvcSnapshotPolUpdateChan <-chan *NetAppVolumeSnapshotPolicyUpdate
	// The channel event handler receives NetAppClusterPeering update events from NVC Client.
	nvcClusterPeeringUpdateChan <-chan *NetAppClusterPeeringUpdate
	// The channel event handler receives NetAppClusterPeering delete events from NVC Client.
	nvcClusterPeeringDeleteChan <-chan *NetAppClusterPeeringDelete
	// The channel event handler receives NetAppQuark update events from NVC Client.
	quarkUpdateChan <-chan *NetAppQuarkUpdate
	// The channel event handler receives NetAppQuark delete events from NVC Client.
	quarkDeleteChan <-chan *NetAppQuarkDelete
	// The channel event handler receives NetAppVolumeReplication update events from NVC Client.
	volReplUpdateChan <-chan *NetAppVolumeReplicationUpdate
	// The channel event handler receives NetAppVolumeReplication delete events from NVC Client.
	volReplDeleteChan <-chan *NetAppVolumeReplicationDelete

	// The last used subscription ID for volume events.
	volEventID int64
	// The last used subscription ID for storagePool events.
	spEventID int64
	// The last used subscription ID for snapshot events.
	snapEventID int64
	// The last used subscription ID for exportPolicy events.
	expEventID int64
	// The last used subscription ID for snapshotPolicy events.
	snapPolicyEventID int64
	// The last used subscription ID for clusterPerring events.
	clusterPeeringEventID int64
	// The last used subscription ID for quark events.
	quarkEventID int64
	// The last used subscription ID for volume replication events.
	volReplEventID int64

	// The notification queue for NetAppVolumes.
	volSubscriptions map[int64]*NetAppVolumeEvent
	// The notification queue for NetAppStoragePool.
	spSubscriptions map[int64]*NetAppStoragePoolEvent
	// The notification queue for NetAppVolumeSnapshots.
	snapSubscriptions map[int64]*NetAppVolumeSnapshotEvent
	// The notification queue for NetAppVolumeCloudSnapshot.
	cloudSnapSubscriptions map[int64]*NetAppVolumeCloudSnapshotEvent
	// The notification queue for NetAppVolumeExportPolicy.
	exportPolSubscriptions map[int64]*NetAppExportPolicyEvent
	// The notification queue for NetAppVolumeSnapshotPolicy.
	snapshotPolSubscriptions map[int64]*NetAppSnapshotPolicyEvent
	// The notification queue for NetAppClusterPeering.
	clusterPeeringSubscriptions map[int64]*NetAppClusterPeeringEvent
	// The notification queue for Quark.
	quarkSubscriptions map[int64]*NetAppQuarkEvent
	// The notification queue for NetAppVolumeReplication
	volReplSubscriptions map[int64]*NetAppVolumeReplicationEvent
	// Event Handler mutex lock for volumes. (TODO: per-volume locks)
	volMutex sync.Mutex
	// Event Handler mutex lock for storagePool. (TODO: per-volume locks)
	spMutex sync.Mutex
	// Event Handler mutex lock for snapshots. (TODO: per-volume locks)
	snapMutex sync.Mutex
	// Event Handler mutex lock for cloud snapshots.
	cloudSnapMutex sync.Mutex
	// Event Handler mutex lock for export policy.
	expMutex sync.Mutex
	// Event Handler mutex lock for snapshot policy.
	snapPolicyMutex sync.Mutex
	// Event Handler mutex lock for clusterPeering.
	clusterPeeringMutex sync.Mutex
	// Event Handler mutex lock for quark.
	quarkMutex sync.Mutex
	// Event Handler mutex lock for volume replication.
	volReplMutex sync.Mutex
}

func newNVCClientEventHandler(client *NVCClient,
	nvcVolUpdateChan <-chan *NetAppVolumeUpdate,
	nvcVolDeleteChan <-chan *NetAppVolumeDelete,
	nvcSpUpdateChan <-chan *NetAppStoragePoolUpdate,
	nvcSpDeleteChan <-chan *NetAppStoragePoolDelete,
	nvcSnapUpdateChan <-chan *NetAppVolumeSnapshotUpdate,
	nvcSnapDeleteChan <-chan *NetAppVolumeSnapshotDelete,
	nvcSnapRestoreChan <-chan *NetAppVolumeSnapshotRestore,
	nvcSnapsListUpdateChan <-chan *NetAppVolumeSnapshotsListUpdate,
	nvcCloudSnapUpdateChan <-chan *NetAppVolumeCloudSnapshotUpdate,
	nvcCloudSnapDeleteChan <-chan *NetAppVolumeCloudSnapshotDelete,
	nvcExportPolUpdateChan <-chan *NetAppVolumeExportPolicyUpdate,
	nvcSnapshotPolUpdateChan <-chan *NetAppVolumeSnapshotPolicyUpdate,
	quarkUpdateChan <-chan *NetAppQuarkUpdate,
	quarkDeleteChan <-chan *NetAppQuarkDelete,
	volReplUpdateChan <-chan *NetAppVolumeReplicationUpdate,
	volReplDeleteChan <-chan *NetAppVolumeReplicationDelete,
	nvcClusterPeeringUpdateChan <-chan *NetAppClusterPeeringUpdate,
	nvcClusterPeeringDeleteChan <-chan *NetAppClusterPeeringDelete) *NVCClientEventHandler {

	return &NVCClientEventHandler{
		client:                      client,
		nvcVolUpdateChan:            nvcVolUpdateChan,
		nvcVolDeleteChan:            nvcVolDeleteChan,
		nvcSpUpdateChan:             nvcSpUpdateChan,
		nvcSpDeleteChan:             nvcSpDeleteChan,
		nvcSnapUpdateChan:           nvcSnapUpdateChan,
		nvcSnapDeleteChan:           nvcSnapDeleteChan,
		nvcSnapRestoreChan:          nvcSnapRestoreChan,
		nvcSnapsListUpdateChan:      nvcSnapsListUpdateChan,
		nvcCloudSnapUpdateChan:      nvcCloudSnapUpdateChan,
		nvcCloudSnapDeleteChan:      nvcCloudSnapDeleteChan,
		nvcExportPolUpdateChan:      nvcExportPolUpdateChan,
		nvcSnapshotPolUpdateChan:    nvcSnapshotPolUpdateChan,
		quarkUpdateChan:             quarkUpdateChan,
		quarkDeleteChan:             quarkDeleteChan,
		volReplUpdateChan:           volReplUpdateChan,
		volReplDeleteChan:           volReplDeleteChan,
		nvcClusterPeeringUpdateChan: nvcClusterPeeringUpdateChan,
		nvcClusterPeeringDeleteChan: nvcClusterPeeringDeleteChan,
		volEventID:                  1,
		spEventID:                   1,
		snapEventID:                 1,
		expEventID:                  1,
		snapPolicyEventID:           1,
		quarkEventID:                1,
		volReplEventID:              1,
		clusterPeeringEventID:       1,
		volSubscriptions:            make(map[int64]*NetAppVolumeEvent),
		spSubscriptions:             make(map[int64]*NetAppStoragePoolEvent),
		snapSubscriptions:           make(map[int64]*NetAppVolumeSnapshotEvent),
		cloudSnapSubscriptions:      make(map[int64]*NetAppVolumeCloudSnapshotEvent),
		exportPolSubscriptions:      make(map[int64]*NetAppExportPolicyEvent),
		snapshotPolSubscriptions:    make(map[int64]*NetAppSnapshotPolicyEvent),
		quarkSubscriptions:          make(map[int64]*NetAppQuarkEvent),
		volReplSubscriptions:        make(map[int64]*NetAppVolumeReplicationEvent),
		clusterPeeringSubscriptions: make(map[int64]*NetAppClusterPeeringEvent),
		volMutex:                    sync.Mutex{},
		spMutex:                     sync.Mutex{},
		snapMutex:                   sync.Mutex{},
		cloudSnapMutex:              sync.Mutex{},
		expMutex:                    sync.Mutex{},
		snapPolicyMutex:             sync.Mutex{},
		quarkMutex:                  sync.Mutex{},
		volReplMutex:                sync.Mutex{},
		clusterPeeringMutex:         sync.Mutex{},
	}
}

func (h *NVCClientEventHandler) Run(wg *sync.WaitGroup) {
	defer wg.Done()
	// Instead of reuse, we create an new WaitGroup to track internal
	// goroutines.
	handlerWg := sync.WaitGroup{}
	defer handlerWg.Wait()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case volUpdate, ok := <-h.nvcVolUpdateChan:
				if !ok {
					log.Info("NVC Client volume update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":                             volUpdate.Volume.Name,
					"resized":                            volUpdate.Status.SizeUpdated,
					"exportPolicyApplied":                volUpdate.Status.ExportPolicyUpdated,
					"snapshotPolicyApplied":              volUpdate.Status.SnapshotPolicyUpdated,
					"smbSharePropertiesApplied":          volUpdate.Status.SMBSharePropertiesUpdated,
					"smbDNSUpdated":                      volUpdate.Status.SMBDNSUpdated,
					"smbPasswordReset":                   volUpdate.Status.SMBPasswordReset,
					"smbEncryptedADCommunicationUpdated": volUpdate.Status.SMBEncryptedADCommunicationUpdated,
					"NoSnapDirOptionUpdated":             volUpdate.Status.NoSnapDirOptionUpdated,
					"conditions":                         volUpdate.UpdatedConditions,
				}).Debug("Received a NetAppVolume update event.")
				h.volMutex.Lock()
				for id, sub := range h.volSubscriptions {
					if sub.Volume.Name == volUpdate.Volume.Name {
						// First, we pass fatal errors through the event channel.
						// Second, we notify of the subscribed events by closing
						// the channel.
						switch sub.EventType {
						case VolumeOnline:
							if !IsVolumeOnline(volUpdate.UpdatedConditions) {
								// Check if there's a fatal error corresponding to this event.
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg := h.volumeOnlineError(volUpdate.UpdatedConditions)
								if errMsg != "" {
									// only send fatal error on channel
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeOffline:
							if !IsVolumeOnline(volUpdate.UpdatedConditions) {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeResized:
							if !volUpdate.Status.SizeUpdated {
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeResizeError)
								if errMsg != "" {
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										// only send fatal error on channel
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSnapPolicyApplied:
							if !volUpdate.Status.SnapshotPolicyUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeExpPolicyApplied:
							if !volUpdate.Status.ExportPolicyUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBSharePropertiesApplied:
							if !volUpdate.Status.SMBSharePropertiesUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeSMBConfigurationError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBDNSUpdated:
							if volUpdate.Status.SMBDNSUpdated {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}

						case VolumeSMBEncryptedADCommunicationUpdated:
							if volUpdate.Status.SMBEncryptedADCommunicationUpdated {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeNoSnapDirOptionUpdated:
							if volUpdate.Status.NoSnapDirOptionUpdated {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeSMBPasswordReset:
							if !volUpdate.Status.SMBPasswordReset {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeSMBPasswordResetError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						case VolumeVersionUpgraded:
							if !volUpdate.Status.QuarkVersionUpdated {
								errMsg := h.volumeUpdateError(volUpdate.UpdatedConditions, quarkv1alpha1.NetAppVolumeUpgradeFailed)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.volSubscriptions, id)
							}
						}
					}
				}
				h.volMutex.Unlock()
			case vol, ok := <-h.nvcVolDeleteChan:
				if !ok {
					log.Info("NVC Client volume delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume": vol.Volume.Name,
				}).Debug("Received a NetAppVolume delete event.")
				h.volMutex.Lock()
				for id, sub := range h.volSubscriptions {
					if sub.Volume.Name == vol.Volume.Name {
						switch sub.EventType {
						case VolumeDeleted:
							close(sub.Channel)
							delete(h.volSubscriptions, id)
						}
					}
				}
				h.volMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case spUpdate, ok := <-h.nvcSpUpdateChan:
				if !ok {
					log.Info("NVC Client storagePool update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"storagePool":                spUpdate.StoragePool.Name,
					"resized":                    spUpdate.Status.SizeUpdated,
					"smbDNSUpdated":              spUpdate.Status.ADDNSUpdated,
					"smbPasswordReset":           spUpdate.Status.ADPasswordReset,
					"smbADconfigurationUpdated":  spUpdate.Status.SMBADConfigurationUpdated,
					"smbADConfigUserListUpdated": spUpdate.Status.SMBADConfigUserListUpdated,
					"conditions":                 spUpdate.UpdatedConditions,
				}).Debug("Received a NetAppStoragePool update event.")
				h.spMutex.Lock()
				for id, sub := range h.spSubscriptions {
					if sub.StoragePool.Name == spUpdate.StoragePool.Name {
						// First, we pass fatal errors through the event channel.
						// Second, we notify of the subscribed events by closing
						// the channel.
						switch sub.EventType {
						case StoragePoolOnline:
							if !IsStoragePoolOnline(spUpdate.UpdatedConditions) {
								// Check if there's a fatal error corresponding to this event.
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg, forceRetry := h.storagePoolOnlineError(spUpdate.UpdatedConditions)
								if errMsg != "" && !forceRetry {
									// only send fatal error on channel
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolOffline:
							if !IsStoragePoolOnline(spUpdate.UpdatedConditions) {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolResized:
							if !spUpdate.Status.SizeUpdated {
								// this error could be a retriable error which is considered fatal only after timeout
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolResizeError)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										// only send fatal error on channel
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolMigratedZone:
							if !spUpdate.Status.ZoneUpdated {
								errMsg := h.storagePoolMigrateZoneError(
									spUpdate.LastZoneMigrationCause)
								if errMsg != "" {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolGlobalAccessUpdated:
							if !spUpdate.Status.GlobalAccessUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(
									spUpdate.UpdatedConditions,
									quarkv1alpha1.NetAppStoragePoolAddressAllocationError)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolVersionUpgraded:
							if !spUpdate.Status.QuarkVersionUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolUpgradeFailed)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolSMBDNSUpdated:
							if !spUpdate.Status.ADDNSUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolSMBPasswordReset:
							if !spUpdate.Status.ADPasswordReset {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolSMBADConfigUserListUpdated:
							if !spUpdate.Status.SMBADConfigUserListUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolSMBBackupUsersError)
								if errMsg == "" {
									errMsg, forceRetry = h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolSMBSecurityPrivilegeUsersError)
								}
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolLdapConfigured:
							if !spUpdate.Status.LdapConfigured {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
								if errMsg == "" {
									errMsg, forceRetry = h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
								}
								if errMsg == "" {
									errMsg, forceRetry = h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolLdapConfigurationError)
								}
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolSMBADConfigurationUpdated:
							if !spUpdate.Status.SMBADConfigurationUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						case StoragePoolSMBServerConfigurationUpdated:
							if !spUpdate.Status.SMBServerConfigurationUpdated {
								errMsg, forceRetry := h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
								if errMsg == "" {
									errMsg, forceRetry = h.storagePoolUpdateError(spUpdate.UpdatedConditions, quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
								}
								if errMsg != "" && !forceRetry {
									if !util.VerifyRetryErrorMessage(errMsg) {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.spSubscriptions, id)
							}
						}
					}
				}
				h.spMutex.Unlock()
			case sp, ok := <-h.nvcSpDeleteChan:
				if !ok {
					log.Info("NVC Client storagePool delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"storagePool": sp.StoragePool.Name,
				}).Debug("Received a NetAppStoragePool delete event.")
				h.spMutex.Lock()
				for id, sub := range h.spSubscriptions {
					if sub.StoragePool.Name == sp.StoragePool.Name {
						switch sub.EventType {
						case StoragePoolDeleted:
							close(sub.Channel)
							delete(h.spSubscriptions, id)
						}
					}
				}
				h.spMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case snapUpdate, ok := <-h.nvcSnapUpdateChan:
				if !ok {
					log.Info("NVC Client snapshot update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"snapshot":   snapUpdate.Snapshot.Name,
					"conditions": snapUpdate.UpdatedConditions,
				}).Debug("Received a NetAppVolumeSnapshot update event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if sub.SnapDisplayName == snapUpdate.Snapshot.Spec.DisplayName &&
						snapUpdate.Snapshot.ObjectMeta.GetLabels()[quarkconfig.VolumeLabel] == sub.VolumeUUID {
						switch sub.EventType {
						case SnapshotReady:
							if !IsSnapshotReady(snapUpdate.UpdatedConditions) {
								errMsg := h.snapshotError(snapUpdate.UpdatedConditions, quarkv1alpha1.SnapshotCreateError)
								if errMsg != "" {
									sub.Channel <- NewNVCClientFatalError(errMsg)
								}
							} else {
								close(sub.Channel)
								delete(h.snapSubscriptions, id)
							}
						}
					}
				}
				h.snapMutex.Unlock()
			case snapDelete, ok := <-h.nvcSnapDeleteChan:
				if !ok {
					log.Info("NVC Client snapshot delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"deletedSnapshots": snapDelete.DeletedSnapshots,
				}).Debug("Received a NetAppVolumeSnapshot delete event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapDelete.VolumeUUID == sub.VolumeUUID &&
						util.StringSetContains(snapDelete.DeletedSnapshots, sub.SnapDisplayName) {
						switch sub.EventType {
						case SnapshotDeleted:
							close(sub.Channel)
							delete(h.snapSubscriptions, id)
						}
					}
				}
				h.snapMutex.Unlock()
			case snapRestore, ok := <-h.nvcSnapRestoreChan:
				if !ok {
					log.Info("NVC Client snapshot restore event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":           snapRestore.VolumeUUID,
					"errorCondition":   snapRestore.ErrorCondition,
					"restoredSnapshot": snapRestore.RestoredSnapshot,
				}).Debug("Received a NetAppVolumeSnapshot restore event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapRestore.VolumeUUID == sub.VolumeUUID {
						switch sub.EventType {
						case SnapshotRestored:
							if !snapRestore.RestoredSnapshot {
								errMsg := snapRestore.ErrorCondition
								if errMsg != "" {
									if !strings.Contains(strings.ToLower(errMsg), "retrying") {
										sub.Channel <- NewNVCClientFatalError(errMsg)
									}
								}
							} else {
								close(sub.Channel)
								delete(h.snapSubscriptions, id)
							}
						}
					}
				}
				h.snapMutex.Unlock()
			case snapListUpdate, ok := <-h.nvcSnapsListUpdateChan:
				if !ok {
					log.Info("NVC Client snapshotsList update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume":               snapListUpdate.VolumeUUID,
					"snapshotsListUpdated": snapListUpdate.SnapshotsListUpdated,
				}).Debug("Received a NetAppVolumeSnapshotsListUpdate event.")
				h.snapMutex.Lock()
				for id, sub := range h.snapSubscriptions {
					if snapListUpdate.VolumeUUID == sub.VolumeUUID &&
						snapListUpdate.SnapshotsListUpdated {
						switch sub.EventType {
						case SnapshotsListUpdated:
							close(sub.Channel)
							delete(h.snapSubscriptions, id)
						}
					}
				}
				h.snapMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case cloudSnapUpdate, ok := <-h.nvcCloudSnapUpdateChan:
				if !ok {
					log.Info("NVC Client cloud snapshot update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"cloudSnapshot":     cloudSnapUpdate.Status.CloudSnapshotUUID,
					"ready":             cloudSnapUpdate.Status.Ready,
					"uploading":         cloudSnapUpdate.Status.Uploading,
					"deleting":          cloudSnapUpdate.Status.Deleting,
					"endpointUUID":      cloudSnapUpdate.Status.EndpointUUID,
					"bytesTransferred":  cloudSnapUpdate.Status.BytesTransferred,
					"logicalSize":       cloudSnapUpdate.Status.LogicalSize,
					"completionPercent": cloudSnapUpdate.Status.CompletionPercent,
					"error":             cloudSnapUpdate.Status.Error,
					"creationTime":      cloudSnapUpdate.Status.CreationTime,
				}).Debug("Received a NetAppVolumeCloudSnapshot update event.")
				h.volMutex.Lock()
				for id, sub := range h.cloudSnapSubscriptions {
					if sub.CloudSnapshot == cloudSnapUpdate.CloudSnapshot.Name {
						switch sub.EventType {
						case CloudSnapshotReady:
							if cloudSnapUpdate.Status.Ready {
								close(sub.Channel)
								delete(h.cloudSnapSubscriptions, id)
							}
						}
					}
				}
				h.volMutex.Unlock()
			case cloudSnapDelete, ok := <-h.nvcCloudSnapDeleteChan:
				if !ok {
					log.Info("NVC Client cloud snapshot delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"cloudSnapshot": cloudSnapDelete.CloudSnapshot.Name,
				}).Debug("Received a NetAppVolumeCloudSnapshot delete event.")
				h.volMutex.Lock()
				for id, sub := range h.cloudSnapSubscriptions {
					if sub.CloudSnapshot == cloudSnapDelete.CloudSnapshot.Name {
						switch sub.EventType {
						case CloudSnapshotDeleted:
							close(sub.Channel)
							delete(h.cloudSnapSubscriptions, id)
						}
					}
				}
				h.volMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case exportPolicyUpdate, ok := <-h.nvcExportPolUpdateChan:
				if !ok {
					log.Info("NVC Client cloud snapshot update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"exportRules": exportPolicyUpdate.Rules,
				}).Debug("Received a exportPolicyUpdate event.")
				h.expMutex.Lock()
				for id, sub := range h.exportPolSubscriptions {
					if sub.ExportPolicy == exportPolicyUpdate.ExportPolicy.Name {
						switch sub.EventType {
						case ExportPolicyUpdated:
							close(sub.Channel)
							delete(h.exportPolSubscriptions, id)
						}
					}
				}
				h.expMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case snapshotPolicyUpdate, ok := <-h.nvcSnapshotPolUpdateChan:
				if !ok {
					log.Info("NVC Client snapshot policy update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"Enabled":           snapshotPolicyUpdate.Enabled,
					"snapshotSchedules": snapshotPolicyUpdate.Schedules,
				}).Debug("Received a snapshotPolicyUpdate event.")
				h.snapPolicyMutex.Lock()
				for id, sub := range h.snapshotPolSubscriptions {
					if sub.SnapshotPolicy == snapshotPolicyUpdate.SnapshotPolicy.Name {
						switch sub.EventType {
						case SnapshotPolicyUpdated:
							close(sub.Channel)
							delete(h.snapshotPolSubscriptions, id)
						}
					}
				}
				h.snapPolicyMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case quarkUpdate, ok := <-h.quarkUpdateChan:
				if !ok {
					log.Info("NVC Client quark update event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"quark":                 quarkUpdate.Quark.Name,
					"volumeControllerReady": quarkUpdate.Status.VolumeControllerReady,
					"highAvailabilityReady": quarkUpdate.Status.HighAvailabilityReady,
					"operatorReady":         quarkUpdate.Status.OperatorReady,
					"version":               quarkUpdate.Status.Version,
				}).Debug("Received a Quark update event.")
				h.quarkMutex.Lock()
				for id, sub := range h.quarkSubscriptions {
					if sub.QuarkName == quarkUpdate.Quark.Name {
						switch sub.EventType {
						case QuarkReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkNotReady:
							if !quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkOperatorReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.OperatorReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case QuarkOperatorNotReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if !quarkUpdate.Status.OperatorReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case VolumeControllerReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case VolumeControllerNotReady:
							// TODO: Add NVCClientFatalError if needed to channel
							if !quarkUpdate.Status.VolumeControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case HighAvailabilityReady:
							if quarkUpdate.Status.HighAvailabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case HighAvailabilityNotReady:
							if !quarkUpdate.Status.HighAvailabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ExternalVolumeAccessReady:
							if quarkUpdate.Status.ExternalVolumeAccessReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ExternalVolumeAccessNotReady:
							if !quarkUpdate.Status.ExternalVolumeAccessReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case CloudsnapshotCapabilityReady:
							if quarkUpdate.Status.CloudsnapshotCapabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case CloudsnapshotCapabilityNotReady:
							if !quarkUpdate.Status.CloudsnapshotCapabilityReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case IPDReady:
							if quarkUpdate.Status.IPDReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case IPDNotReady:
							if !quarkUpdate.Status.IPDReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case EtcdClusterReady:
							if quarkUpdate.Status.EtcdClusterReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case EtcdClusterNotReady:
							if !quarkUpdate.Status.EtcdClusterReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case NodevolReady:
							if quarkUpdate.Status.NodevolReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case NodevolNotReady:
							if !quarkUpdate.Status.NodevolReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ServiceMeshControllerReady:
							if quarkUpdate.Status.ServiceMeshControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						case ServiceMeshControllerNotReady:
							if !quarkUpdate.Status.ServiceMeshControllerReady {
								close(sub.Channel)
								delete(h.quarkSubscriptions, id)
							}
						}
					}
				}
				h.quarkMutex.Unlock()
			case quarkDelete, ok := <-h.quarkDeleteChan:
				if !ok {
					log.Info("NVC Client quark delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"quark": quarkDelete.Quark.Name,
				}).Debug("Received a Quark delete event.")
				h.quarkMutex.Lock()
				for id, sub := range h.quarkSubscriptions {
					if sub.QuarkName == quarkDelete.Quark.Name {
						switch sub.EventType {
						case QuarkDeleted:
							// TODO: Add NVCClientFatalError if needed to channel
							close(sub.Channel)
							delete(h.quarkSubscriptions, id)
						}
					}
				}
				h.quarkMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case volReplUpdate, ok := <-h.volReplUpdateChan:
				if !ok {
					log.Info("NVC Client NetAppVolumeReplication update event handler has terminated")
					return
				}
				log.WithFields(log.Fields{
					"volumeReplication":          volReplUpdate.VolumeReplication.Name,
					"mirrorStateUpdated":         volReplUpdate.Status.MirrorStateUpdated,
					"replicationScheduleUpdated": volReplUpdate.Status.ReplicationScheduleUpdated,
					"replicationStatusUpdated":   volReplUpdate.Status.ReplicationStatusUpdated,
					"conditions":                 volReplUpdate.UpdatedConditions,
				}).Debug("Received NetAppVolumeReplication update event.")
				h.volReplMutex.Lock()
				for id, sub := range h.volReplSubscriptions {
					if sub.VolumeReplication.Name == volReplUpdate.VolumeReplication.Name {
						switch sub.EventType {
						case VolumeReplicationInitialized:
							if volReplUpdate.Status.MirrorStateUpdated &&
								NetAppVolumeReplicationMirrored(volReplUpdate.VolumeReplication) {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						case VolumeReplicationBroken:
							if volReplUpdate.Status.MirrorStateUpdated &&
								NetAppVolumeReplicationBroken(volReplUpdate.VolumeReplication) {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						case VolumeReplicationHealthy:
							if NetAppVolumeReplicationHealthy(volReplUpdate.VolumeReplication) {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						case VolumeReplicationScheduleUpdated:
							if volReplUpdate.Status.ReplicationScheduleUpdated {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						case VolumeReplicationStatusIdle:
							if volReplUpdate.Status.ReplicationStatusUpdated &&
								NetAppVolumeReplicationStatus(volReplUpdate.VolumeReplication) ==
									quarkv1alpha1.ReplicationStatusIdle {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						case VolumeReplicationStatusTransferring:
							if volReplUpdate.Status.ReplicationStatusUpdated &&
								NetAppVolumeReplicationStatus(volReplUpdate.VolumeReplication) ==
									quarkv1alpha1.ReplicationStatusTransferring {
								close(sub.Channel)
								delete(h.volReplSubscriptions, id)
							}
						}
					}
				}
				h.volReplMutex.Unlock()
			case volReplDelete, ok := <-h.volReplDeleteChan:
				if !ok {
					log.Info("NVC Client NetAppVolumeReplication delete event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"volume": volReplDelete.VolumeReplication.Name,
				}).Debug("Received a NetAppVolumeReplication delete event.")
				h.volReplMutex.Lock()
				for id, sub := range h.volReplSubscriptions {
					if sub.VolumeReplication.Name == volReplDelete.VolumeReplication.Name {
						switch sub.EventType {
						case VolumeReplicationDeleted:
							close(sub.Channel)
							delete(h.volReplSubscriptions, id)
						}
					}
				}
				h.volReplMutex.Unlock()
			}
		}
	}()
	go func() {
		defer handlerWg.Done()
		for {
			select {
			case clusterPeeringUpdate, ok := <-h.nvcClusterPeeringUpdateChan:
				if !ok {
					log.Info("NVC Client NetAppClusterPeering event handler has terminated")
					return
				}
				log.WithFields(log.Fields{
					"clusterPeering": clusterPeeringUpdate.ClusterPeering.Name,
					"conditions":     clusterPeeringUpdate.UpdatedConditions,
				}).Debug("Received NetAppClusterPeering update event.")
				h.clusterPeeringMutex.Lock()
				for id, sub := range h.clusterPeeringSubscriptions {
					if sub.ClusterPeering.Name == clusterPeeringUpdate.ClusterPeering.Name {
						switch sub.EventType {
						case ClusterPeeringReady:
							if NetAppClusterPeeringReady(clusterPeeringUpdate.ClusterPeering) {
								close(sub.Channel)
								delete(h.clusterPeeringSubscriptions, id)
							}
						case ServiceMeshEnabled:
							if NetAppServiceMeshEnabled(clusterPeeringUpdate.ClusterPeering) {
								close(sub.Channel)
								delete(h.clusterPeeringSubscriptions, id)
							}
						}
					}
				}
				h.clusterPeeringMutex.Unlock()
			case clusterPeeringDelete, ok := <-h.nvcClusterPeeringDeleteChan:
				if !ok {
					log.Info("NVC Client NetAppClusterPeering event handler has terminated.")
					return
				}
				log.WithFields(log.Fields{
					"clusterPeering": clusterPeeringDelete.ClusterPeering.Name,
				}).Debug("Received a NetAppClusterPeering delete event.")
				h.clusterPeeringMutex.Lock()
				for id, sub := range h.clusterPeeringSubscriptions {
					if sub.ClusterPeering.Name == clusterPeeringDelete.ClusterPeering.Name {
						switch sub.EventType {
						case ClusterPeeringDeleted:
							close(sub.Channel)
							delete(h.clusterPeeringSubscriptions, id)
						}
					}
				}
				h.clusterPeeringMutex.Unlock()
			}
		}
	}()
	handlerWg.Add(1)
}

func (h *NVCClientEventHandler) SubscribeVolumeEvent(event *NetAppVolumeEvent) int64 {
	h.volMutex.Lock()
	defer h.volMutex.Unlock()
	vol, err := h.client.NetAppVolumeGet(event.Volume.Name)
	if vol != nil && err == nil {
		switch event.EventType {
		case VolumeOnline:
			if h.client.NetAppVolumeOnline(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeOffline:
			if !h.client.NetAppVolumeOnline(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeResized:
			if h.client.NetAppVolumeResized(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeExpPolicyApplied:
			if h.client.NetAppVolExportPolicyUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSnapPolicyApplied:
			if h.client.NetAppVolSnapPolicyUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeNoSnapDirOptionUpdated:
			if h.client.NetAppVolNoSnapDirOptionUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBSharePropertiesApplied:
			if h.client.NetAppVolSMBSharePropertiesUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBDNSUpdated:
			if h.client.NetAppVolSMBDNSUpdated(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBPasswordReset:
			if h.client.NetAppVolSMBPasswordResetDone(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeVersionUpgraded:
			if h.client.NetAppVolumeUpgraded(vol) {
				close(event.Channel)
				return 0
			}
		case VolumeSMBEncryptedADCommunicationUpdated:
			if h.client.NetAppVolSMBEncryptedADCommunicationUpdated(vol) {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case VolumeDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.volSubscriptions[h.volEventID] = event
	ret := h.volEventID
	h.volEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeVolumeEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.volMutex.Lock()
	close(h.volSubscriptions[eventID].Channel)
	delete(h.volSubscriptions, eventID)
	h.volMutex.Unlock()
}

// WaitVolumeEventWithTimeout is a blocking call that returns when a volume
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitVolumeEventWithTimeout(
	volume *quarkv1alpha1.NetAppVolume,
	eventType VolumeEvent,
	timeout time.Duration) error {
	volEvent := &NetAppVolumeEvent{
		Volume:    volume,
		EventType: eventType,
		Channel:   make(chan error),
	}
	// return fatal error if occurred before subscribing
	retryErr, err := h.processVolumeEventError(volume.Name, eventType)
	if err != nil && !retryErr {
		return err
	}
	subID := h.SubscribeVolumeEvent(volEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-volEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeVolumeEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeVolumeEvent(subID)
				_, err := h.processVolumeEventError(volume.Name, eventType)
				if err != nil {
					return err
				}
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) WaitVolumeForStoragepoolOnlineEventWithTimeout(
	storagePoolUUID string,
	timeout time.Duration) (string, error) {

	volumes, err := h.client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return "", err
	}
	for _, volume := range volumes.Items {
		err = h.WaitVolumeEventWithTimeout(&volume,
			VolumeOnline, timeout)
		if err != nil {
			return volume.Name, err
		}
	}
	return "", nil
}

func (h *NVCClientEventHandler) SubscribeStoragePoolEvent(event *NetAppStoragePoolEvent) int64 {
	h.spMutex.Lock()
	defer h.spMutex.Unlock()
	sp, err := h.client.NetAppStoragePoolGet(event.StoragePool.Name)
	if sp != nil && err == nil {
		switch event.EventType {
		case StoragePoolOnline:
			if h.client.NetAppStoragePoolOnline(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolOffline:
			if !h.client.NetAppStoragePoolOnline(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolResized:
			if h.client.NetAppStoragePoolResized(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolMigratedZone:
			if h.client.NetAppStoragePoolMigratedZone(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolGlobalAccessUpdated:
			if h.client.NetAppStoragePoolGlobalAccessUpdated(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolVersionUpgraded:
			if h.client.NetAppStoragePoolUpgraded(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolSMBDNSUpdated:
			if h.client.NetAppStoragePoolADDNSUpdated(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolSMBPasswordReset:
			if h.client.NetAppStoragePoolADPasswordResetDone(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolSMBADConfigurationUpdated:
			if h.client.NetAppStoragePoolSMBADConfigurationUpdated(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolSMBADConfigUserListUpdated:
			if h.client.NetAppStoragePoolSMBADUserListUpdated(sp) {
				close(event.Channel)
				return 0
			}
		case StoragePoolLdapConfigured:
			if h.client.NetAppStoragePoolLdapReady(sp) {
				close(event.Channel)
				return 0
			}

		case StoragePoolSMBServerConfigurationUpdated:
			if h.client.NetAppStoragePoolSMBServerConfigurationUpdated(sp) {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case StoragePoolDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.spSubscriptions[h.spEventID] = event
	ret := h.spEventID
	h.spEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeStoragePoolEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.spMutex.Lock()
	close(h.spSubscriptions[eventID].Channel)
	delete(h.spSubscriptions, eventID)
	h.spMutex.Unlock()
}

// WaitStoragePoolEventWithTimeout is a blocking call that returns when a storagePool
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitStoragePoolEventWithTimeout(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	eventType StoragePoolEvent,
	timeout time.Duration) error {
	spEvent := &NetAppStoragePoolEvent{
		StoragePool: storagePool,
		EventType:   eventType,
		Channel:     make(chan error),
	}
	// return fatal error if occurred before subscribing
	retryErr, err := h.processStoragePoolEventError(storagePool.Name, eventType)
	if err != nil && !retryErr {
		return err
	}
	subID := h.SubscribeStoragePoolEvent(spEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-spEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeStoragePoolEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				// Note that if a zone fails between processing the storage pool event
				// and subscribing, this will likely hit the timeout; it won't
				// erroneously return success, however.
				h.UnsubscribeStoragePoolEvent(subID)
				retryErr, err := h.processStoragePoolEventError(storagePool.Name, eventType)
				eventTimeoutString := getStoragePoolEventTimeOutString(eventType, timeout)
				if err != nil {
					// DNS IP and Domain unreachable errors should be fatal errors, so that it will be propogated to end user
					if !retryErr || isADServerError(err.Error()) {
						return err
					}
					return NewNVCClientTimeoutErrorStoragePool(eventTimeoutString)
				} else {
					return NewNVCClientTimeoutErrorStoragePool(eventTimeoutString)
				}
			}
		}
	}
	return nil
}

func isADServerError(errMsg string) bool {
	return strings.Contains(strings.ToLower(errMsg), strings.ToLower(quarkconfig.MessageADServerError))
}

func getStoragePoolEventTimeOutString(eventType StoragePoolEvent, timeout time.Duration) string {
	switch eventType {
	case StoragePoolOnline:
		return fmt.Sprintf("Failed to bring StoragePool online in %s", timeout)
	case StoragePoolResized:
		return fmt.Sprintf("Failed to resize StoragePool in %s", timeout)
	case StoragePoolSMBDNSUpdated:
		return fmt.Sprintf("Failed to configure DNS in %s", timeout)
	case StoragePoolSMBADConfigUserListUpdated:
		return fmt.Sprintf("Failed to configure SMB AD user list in %s", timeout)
	case StoragePoolGlobalAccessUpdated:
		return fmt.Sprintf("Failed to configure Global Access in %s", timeout)
	case StoragePoolSMBADConfigurationUpdated:
		return fmt.Sprintf("Failed to configure SMB AD in %s", timeout)
	case StoragePoolSMBServerConfigurationUpdated:
		return fmt.Sprintf("Failed to configure SMB server in %s", timeout)
	case StoragePoolSMBPasswordReset:
		return fmt.Sprintf("Failed to update SMB password in %s", timeout)
	case StoragePoolVersionUpgraded:
		return fmt.Sprintf("Failed to upgrade StoragePool version in %s", timeout)
	case StoragePoolMigratedZone:
		return fmt.Sprintf("Failed to migrate StoragePool in %s", timeout)
	default:
		return fmt.Sprintf("Invalid StoragePool event")
	}
}

func (h *NVCClientEventHandler) SubscribeQuarkEvent(event *NetAppQuarkEvent) int64 {
	h.quarkMutex.Lock()
	defer h.quarkMutex.Unlock()
	quark, err := h.client.NetAppQuarkGet(event.QuarkName)
	if quark != nil && err == nil {
		switch event.EventType {
		case QuarkReady:
			if h.client.NetAppQuarkReady(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkNotReady:
			if !h.client.NetAppQuarkReady(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkOperatorReady:
			if h.client.NetAppQuarkOperatorReadyforEvent(quark) {
				close(event.Channel)
				return 0
			}
		case QuarkOperatorNotReady:
			if !h.client.NetAppQuarkOperatorReadyforEvent(quark) {
				close(event.Channel)
				return 0
			}
		case VolumeControllerReady:
			if status := getNetAppQuarkStatus(quark); status.VolumeControllerReady {
				close(event.Channel)
				return 0
			}
		case VolumeControllerNotReady:
			if status := getNetAppQuarkStatus(quark); !status.VolumeControllerReady {
				close(event.Channel)
				return 0
			}
		case HighAvailabilityReady:
			if status := getNetAppQuarkStatus(quark); status.HighAvailabilityReady {
				close(event.Channel)
				return 0
			}
		case HighAvailabilityNotReady:
			if status := getNetAppQuarkStatus(quark); !status.HighAvailabilityReady {
				close(event.Channel)
				return 0
			}
		case ExternalVolumeAccessReady:
			if status := getNetAppQuarkStatus(quark); status.ExternalVolumeAccessReady {
				close(event.Channel)
				return 0
			}
		case ExternalVolumeAccessNotReady:
			if status := getNetAppQuarkStatus(quark); !status.ExternalVolumeAccessReady {
				close(event.Channel)
				return 0
			}
		case CloudsnapshotCapabilityReady:
			if status := getNetAppQuarkStatus(quark); status.CloudsnapshotCapabilityReady {
				close(event.Channel)
				return 0
			}
		case CloudsnapshotCapabilityNotReady:
			if status := getNetAppQuarkStatus(quark); !status.CloudsnapshotCapabilityReady {
				close(event.Channel)
				return 0
			}
		case IPDReady:
			if status := getNetAppQuarkStatus(quark); status.IPDReady {
				close(event.Channel)
				return 0
			}
		case IPDNotReady:
			if status := getNetAppQuarkStatus(quark); !status.IPDReady {
				close(event.Channel)
				return 0
			}
		case EtcdClusterReady:
			if status := getNetAppQuarkStatus(quark); status.EtcdClusterReady {
				close(event.Channel)
				return 0
			}
		case EtcdClusterNotReady:
			if status := getNetAppQuarkStatus(quark); !status.EtcdClusterReady {
				close(event.Channel)
				return 0
			}
		case NodevolReady:
			if status := getNetAppQuarkStatus(quark); status.NodevolReady {
				close(event.Channel)
				return 0
			}
		case NodevolNotReady:
			if status := getNetAppQuarkStatus(quark); !status.NodevolReady {
				close(event.Channel)
				return 0
			}
		case ServiceMeshControllerReady:
			if status := getNetAppQuarkStatus(quark); status.ServiceMeshControllerReady {
				close(event.Channel)
				return 0
			}
		case ServiceMeshControllerNotReady:
			if status := getNetAppQuarkStatus(quark); !status.ServiceMeshControllerReady {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case QuarkDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.quarkSubscriptions[h.quarkEventID] = event
	ret := h.quarkEventID
	h.quarkEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeQuarkEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.quarkMutex.Lock()
	close(h.quarkSubscriptions[eventID].Channel)
	delete(h.quarkSubscriptions, eventID)
	h.quarkMutex.Unlock()
}

// WaitQuarkEventWithTimeout is a blocking call that returns when a quark
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitQuarkEventWithTimeout(
	quarkName string,
	eventType QuarkEvent,
	timeout time.Duration) error {
	quarkEvent := &NetAppQuarkEvent{
		QuarkName: quarkName,
		EventType: eventType,
		Channel:   make(chan error),
	}

	subID := h.SubscribeQuarkEvent(quarkEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-quarkEvent.Channel:
				if ok {
					// TODO: Return NVCClientFatalError if raised
					h.UnsubscribeQuarkEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				h.UnsubscribeQuarkEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeSnapshotEvent(event *NetAppVolumeSnapshotEvent) int64 {
	h.snapMutex.Lock()
	defer h.snapMutex.Unlock()
	switch event.EventType {
	case SnapshotReady:
		if h.client.NetAppVolumeSnapshotReady(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	case SnapshotDeleted:
		if h.client.NetAppVolumeSnapshotDeleted(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	case SnapshotRestored:
		if h.client.NetAppVolumeSnapshotRestored(event.VolumeUUID, event.SnapDisplayName) {
			close(event.Channel)
			return 0
		}
	}

	h.snapSubscriptions[h.snapEventID] = event
	ret := h.snapEventID
	h.snapEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeSnapshotEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.snapMutex.Lock()
	close(h.snapSubscriptions[eventID].Channel)
	delete(h.snapSubscriptions, eventID)
	h.snapMutex.Unlock()
}

// WaitSnapshotEventWithTimeout is a blocking call that returns when a snapshot
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitSnapshotEventWithTimeout(
	volName string, snapDisplayName string,
	eventType SnapshotEvent,
	timeout time.Duration) error {
	snapEvent := &NetAppVolumeSnapshotEvent{
		VolumeUUID:      volName,
		SnapDisplayName: snapDisplayName,
		EventType:       eventType,
		Channel:         make(chan error),
	}
	// return fatal error if occurred before subscribing
	retryErr, err := h.processSnapshotEventError(volName, eventType)
	if err != nil && !retryErr {
		return err
	}
	subID := h.SubscribeSnapshotEvent(snapEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-snapEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeSnapshotEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeSnapshotEvent(subID)
				_, err := h.processSnapshotEventError(volName, eventType)
				if err != nil {
					return err
				}
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}
func (h *NVCClientEventHandler) WaitCloudSnapshotEventWithTimeout(
	cloudSnapshotName string,
	eventType CloudSnapshotEvent,
	timeout time.Duration) error {
	cloudSnapEvent := &NetAppVolumeCloudSnapshotEvent{
		CloudSnapshot: cloudSnapshotName,
		EventType:     eventType,
		Channel:       make(chan error),
	}
	// return fatal error if occurred before subscribing
	subID := h.SubscribeCloudSnapshotEvent(cloudSnapEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-cloudSnapEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeCloudSnapshotEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeCloudSnapshotEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeCloudSnapshotEvent(event *NetAppVolumeCloudSnapshotEvent) int64 {
	h.cloudSnapMutex.Lock()
	defer h.cloudSnapMutex.Unlock()
	switch event.EventType {
	case CloudSnapshotReady:
		if h.client.NetAppVolumeCloudSnapshotReady(event.CloudSnapshot) {
			close(event.Channel)
			return 0
		}
	case CloudSnapshotDeleted:
		if h.client.NetAppVolumeCloudSnapshotDeleted(event.CloudSnapshot) {
			close(event.Channel)
			return 0
		}
	}

	h.cloudSnapSubscriptions[h.snapEventID] = event
	ret := h.snapEventID
	h.snapEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeCloudSnapshotEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.cloudSnapMutex.Lock()
	close(h.cloudSnapSubscriptions[eventID].Channel)
	delete(h.cloudSnapSubscriptions, eventID)
	h.cloudSnapMutex.Unlock()
}

func (h *NVCClientEventHandler) SubscribeExportPolicyEvent(event *NetAppExportPolicyEvent) int64 {
	h.expMutex.Lock()
	defer h.expMutex.Unlock()
	exportPolicy, err := h.client.NetAppExportPolicyGet(event.ExportPolicy)
	if exportPolicy != nil && err == nil {
		switch event.EventType {
		case ExportPolicyUpdated:
			if h.client.NetAppVolumeExportPolicyUpdated(exportPolicy) {
				close(event.Channel)
				return 0
			}
		}
	}
	h.exportPolSubscriptions[h.expEventID] = event
	ret := h.expEventID
	h.expEventID++
	return ret
}

func (h *NVCClientEventHandler) UnsubscribeExportPolicyEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.expMutex.Lock()
	close(h.exportPolSubscriptions[eventID].Channel)
	delete(h.exportPolSubscriptions, eventID)
	h.expMutex.Unlock()
}

// WaitExportPolicyEventWithTimeout is a blocking call that returns when a exportPolicy
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitExportPolicyEventWithTimeout(
	exportPolicy *quarkv1alpha1.NetAppExportPolicy,
	eventType ExportPolicyEvent,
	timeout time.Duration) error {
	expEvent := &NetAppExportPolicyEvent{
		ExportPolicy: exportPolicy.Name,
		EventType:    eventType,
		Channel:      make(chan error),
	}

	subID := h.SubscribeExportPolicyEvent(expEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-expEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeExportPolicyEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeExportPolicyEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

func (h *NVCClientEventHandler) SubscribeSnapshotPolicyEvent(event *NetAppSnapshotPolicyEvent) int64 {
	h.snapPolicyMutex.Lock()
	defer h.snapPolicyMutex.Unlock()
	snapshotPolicy, err := h.client.NetAppSnapshotPolicyGet(event.SnapshotPolicy)
	if snapshotPolicy != nil && err == nil {
		switch event.EventType {
		case SnapshotPolicyUpdated:
			if h.client.NetAppVolumeSnapshotPolicyUpdated(snapshotPolicy) {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case SnapshotPolicyUpdated:
			close(event.Channel)
			return 0
		}
	}
	h.snapshotPolSubscriptions[h.snapPolicyEventID] = event
	ret := h.snapPolicyEventID
	h.snapPolicyEventID++
	return ret
}
func (h *NVCClientEventHandler) UnsubscribeSnapshotPolicyEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.snapPolicyMutex.Lock()
	close(h.snapshotPolSubscriptions[eventID].Channel)
	delete(h.snapshotPolSubscriptions, eventID)
	h.snapPolicyMutex.Unlock()
}

// WaitSnapshotPolicyEventWithTimeout is a blocking call that returns when a snapshotPolicy
// event of interest is seen, when the event handler times out, or when the
// operation that relates to the event of interest has encountered a fatal
// failure.
func (h *NVCClientEventHandler) WaitSnapshotPolicyEventWithTimeout(
	snapshotPolicy *quarkv1alpha1.NetAppSnapshotPolicy,
	eventType SnapshotPolicyEvent,
	timeout time.Duration) error {
	snapPolicyEvent := &NetAppSnapshotPolicyEvent{
		SnapshotPolicy: snapshotPolicy.Name,
		EventType:      eventType,
		Channel:        make(chan error),
	}
	subID := h.SubscribeSnapshotPolicyEvent(snapPolicyEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-snapPolicyEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeSnapshotPolicyEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeSnapshotPolicyEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

// SubscribeClusterPeeringEvent allows to subscribe for ClusterPeeringEvent event
func (h *NVCClientEventHandler) SubscribeClusterPeeringEvent(event *NetAppClusterPeeringEvent) int64 {
	h.clusterPeeringMutex.Lock()
	defer h.clusterPeeringMutex.Unlock()

	clusterPeering, err := h.client.NetAppClusterPeeringGet(event.ClusterPeering.Name)
	if clusterPeering != nil && err == nil {
		switch event.EventType {
		case ClusterPeeringReady:
			if NetAppClusterPeeringReady(clusterPeering) {
				close(event.Channel)
				return 0
			}
		case ServiceMeshEnabled:
			if NetAppServiceMeshEnabled(clusterPeering) {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case ClusterPeeringDeleted:
			close(event.Channel)
			return 0
		}
	}

	h.clusterPeeringSubscriptions[h.clusterPeeringEventID] = event
	ret := h.clusterPeeringEventID
	h.clusterPeeringEventID++
	return ret
}

// UnsubscribeClusterPeeringEvent unsubscribes from specific ClusterPeering event identified by id
func (h *NVCClientEventHandler) UnsubscribeClusterPeeringEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.clusterPeeringMutex.Lock()
	defer h.clusterPeeringMutex.Unlock()
	close(h.clusterPeeringSubscriptions[eventID].Channel)
	delete(h.clusterPeeringSubscriptions, eventID)
}

// WaitClusterPeeringWithTimeout subscribes and waits (until timeout) for specific ClusterPeeringEvent
func (h *NVCClientEventHandler) WaitClusterPeeringWithTimeout(
	clusterPeering *quarkv1alpha1.NetAppClusterPeering,
	eventType ClusterPeeringEvent,
	timeout time.Duration) error {
	clusterPeeringEvent := &NetAppClusterPeeringEvent{
		ClusterPeering: clusterPeering,
		EventType:      eventType,
		Channel:        make(chan error),
	}
	subID := h.SubscribeClusterPeeringEvent(clusterPeeringEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-clusterPeeringEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeClusterPeeringEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeClusterPeeringEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

// SubscribeVolumeReplicationEvent allows to subscribe for VolumeReplicationEvent event
func (h *NVCClientEventHandler) SubscribeVolumeReplicationEvent(event *NetAppVolumeReplicationEvent) int64 {
	h.volReplMutex.Lock()
	defer h.volReplMutex.Unlock()

	volRepl, err := h.client.NetAppVolumeReplicationGet(event.VolumeReplication.Name)
	if volRepl != nil && err == nil {
		switch event.EventType {
		case VolumeReplicationInitialized:
			if NetAppVolumeReplicationMirrored(volRepl) {
				close(event.Channel)
				return 0
			}
		case VolumeReplicationBroken:
			if NetAppVolumeReplicationBroken(volRepl) {
				close(event.Channel)
				return 0
			}
		case VolumeReplicationHealthy:
			if NetAppVolumeReplicationHealthy(volRepl) {
				close(event.Channel)
				return 0
			}
		case VolumeReplicationScheduleUpdated:
			if NetAppVolumeReplicationScheduleUpdated(volRepl) {
				close(event.Channel)
				return 0
			}
		case VolumeReplicationStatusIdle:
			if NetAppVolumeReplicationStatus(volRepl) ==
				quarkv1alpha1.ReplicationStatusIdle {
				close(event.Channel)
				return 0
			}
		case VolumeReplicationStatusTransferring:
			if NetAppVolumeReplicationStatus(volRepl) ==
				quarkv1alpha1.ReplicationStatusTransferring {
				close(event.Channel)
				return 0
			}
		}
	} else {
		switch event.EventType {
		case VolumeReplicationDeleted:
			close(event.Channel)
			return 0
		}
	}
	h.volReplSubscriptions[h.volReplEventID] = event
	ret := h.volReplEventID
	h.volReplEventID++
	return ret

}

// UnsubscribeVolumeReplicationEvent unsubscribes from specific VolumeReplicationEvent event identified by id
func (h *NVCClientEventHandler) UnsubscribeVolumeReplicationEvent(eventID int64) {
	if eventID == 0 {
		return
	}
	h.volReplMutex.Lock()
	defer h.volReplMutex.Unlock()
	close(h.volReplSubscriptions[eventID].Channel)
	delete(h.volReplSubscriptions, eventID)
}

// WaitVolumeReplicationWithTimeout subscribes and waits (until timeout) for specific VolumeReplicationEvent
func (h *NVCClientEventHandler) WaitVolumeReplicationWithTimeout(
	volumeReplication *quarkv1alpha1.NetAppVolumeReplication,
	eventType VolumeReplicationEvent,
	timeout time.Duration) error {
	volumeReplicationEvent := &NetAppVolumeReplicationEvent{
		VolumeReplication: volumeReplication,
		EventType:         eventType,
		Channel:           make(chan error),
	}
	subID := h.SubscribeVolumeReplicationEvent(volumeReplicationEvent)
	if subID > 0 {
		for {
			select {
			case err, ok := <-volumeReplicationEvent.Channel:
				if ok {
					// unsubscribe when fatal error occurred
					h.UnsubscribeVolumeReplicationEvent(subID)
					return err
				} else {
					return nil
				}
			case <-time.After(timeout):
				// unsubscribe upon timeout
				h.UnsubscribeVolumeReplicationEvent(subID)
				return NewNVCClientTimeoutError(string(eventType))
			}
		}
	}
	return nil
}

// IsVolumeOnline returns true if a volume is online.
func IsVolumeOnline(conditions []quarkv1alpha1.NetAppVolumeCondition) bool {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// IsStoragePoolOnline returns true if a storagePool is Online.
func IsStoragePoolOnline(conditions []quarkv1alpha1.NetAppStoragePoolCondition) bool {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// IsSnapshotReady returns true if a snapshot is ready.
func IsSnapshotReady(conditions []quarkv1alpha1.NetAppVolumeSnapshotCondition) bool {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.SnapshotReady && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// processes error if present in NetAppVolume conditions
func (h *NVCClientEventHandler) processVolumeEventError(volumeUUID string, eventType VolumeEvent) (retryErr bool, ret error) {
	var errMsg string
	conditions, _ := h.client.NetAppVolumeGetConditions(volumeUUID)
	if conditions != nil {
		switch eventType {
		case VolumeOnline:
			errMsg = h.volumeOnlineError(conditions)
		case VolumeResized:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeResizeError)
		case VolumeSMBSharePropertiesApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeSMBConfigurationError)
		case VolumeSMBDNSUpdated:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeSMBConfigurationError)
		case VolumeExpPolicyApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeExportPolicyError)
		case VolumeSnapPolicyApplied:
			errMsg = h.volumeUpdateError(conditions, quarkv1alpha1.NetAppVolumeSnapshotPolicyError)
		}
	}
	if errMsg != "" {
		ret, retryErr = createFatalError(errMsg)
	}
	return retryErr, ret
}

// processes error if present in NetAppStoragePool conditions
func (h *NVCClientEventHandler) processStoragePoolEventError(storagePoolUUID string, eventType StoragePoolEvent) (retryErr bool, ret error) {
	var errMsg string
	forceRetry := false
	sp, err := h.client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return true, fmt.Errorf("unable to determine error; failed to get storage "+
			"pool: %v", err)
	}
	conditions := sp.Status.Conditions
	if conditions != nil {
		switch eventType {
		case StoragePoolOnline:
			errMsg, forceRetry = h.storagePoolOnlineError(conditions)
		case StoragePoolResized:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolResizeError)
		case StoragePoolSMBDNSUpdated:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
		case StoragePoolSMBADConfigUserListUpdated:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolSMBBackupUsersError)
			if errMsg == "" {
				errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolSMBSecurityPrivilegeUsersError)
			}
		case StoragePoolLdapConfigured:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
			if errMsg == "" {
				errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
			}
			if errMsg == "" {
				errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolLdapConfigurationError)
			}
		case StoragePoolGlobalAccessUpdated:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolAddressAllocationError)
		case StoragePoolSMBADConfigurationUpdated:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
		case StoragePoolSMBServerConfigurationUpdated:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
			if errMsg == "" {
				errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
			}
		case StoragePoolSMBPasswordReset:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
		case StoragePoolVersionUpgraded:
			errMsg, forceRetry = h.storagePoolUpdateError(conditions, quarkv1alpha1.NetAppStoragePoolUpgradeFailed)
		}
	}
	if eventType == StoragePoolMigratedZone {
		errMsg = h.storagePoolMigrateZoneError(sp.Spec.LastZoneMigrationCause)
	}
	if errMsg != "" && !forceRetry {
		ret, retryErr = createFatalError(errMsg)
	}
	retryErr = retryErr || forceRetry
	return retryErr, ret
}

// processes error if present in NetAppVolumeSnapshot or NetAppVolumeSnapshotsList conditions
func (h *NVCClientEventHandler) processSnapshotEventError(volumeUUID string, eventType SnapshotEvent) (retryErr bool, ret error) {
	var errMsg string
	switch eventType {
	case SnapshotRestored:
		conditions, _ := h.client.NetAppVolumeSnapshotsListGetConditions(volumeUUID)
		if conditions != nil {
			errMsg = h.snapRestoreError(conditions)
		}
	}
	if errMsg != "" {
		ret, retryErr = createFatalError(errMsg)
	}
	return retryErr, ret
}

// returns volume online error message
func (h *NVCClientEventHandler) volumeOnlineError(conditions []quarkv1alpha1.NetAppVolumeCondition) string {
	for _, cond := range conditions {
		if util.StringSetContains(getFatalVolumeOnlineErrors(), string(cond.Type)) && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

func (h *NVCClientEventHandler) volumeUpdateError(conditions []quarkv1alpha1.NetAppVolumeCondition, errType quarkv1alpha1.NetAppVolumeConditionType) string {
	for _, cond := range conditions {
		if cond.Type == errType && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// returns storagePool online error message
func (h *NVCClientEventHandler) storagePoolOnlineError(conditions []quarkv1alpha1.NetAppStoragePoolCondition) (string, bool) {
	for _, cond := range conditions {
		if util.StringSetContains(getFatalStoragePoolOnlineErrors(), string(cond.Type)) && cond.Status == corev1.ConditionTrue {
			return cond.Message, cond.ForceRetry
		}
	}
	return "", false
}

func (h *NVCClientEventHandler) storagePoolUpdateError(conditions []quarkv1alpha1.NetAppStoragePoolCondition, errType quarkv1alpha1.NetAppStoragePoolConditionType) (string, bool) {
	for _, cond := range conditions {
		if cond.Type == errType && cond.Status == corev1.ConditionTrue {
			return cond.Message, cond.ForceRetry
		}
	}
	return "", false
}

func (h *NVCClientEventHandler) storagePoolMigrateZoneError(
	zoneMigrationCause string,
) string {
	if zoneMigrationCause == quarkconfig.ZoneMigrationCauseFailure {
		return "Zone failure"
	}
	return ""
}

// returns snap restore error message
func (h *NVCClientEventHandler) snapRestoreError(conditions []quarkv1alpha1.NetAppVolumeSnapshotsListCondition) string {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.SnapshotRestoreError && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// returns snapshot related error message
func (h *NVCClientEventHandler) snapshotError(conditions []quarkv1alpha1.NetAppVolumeSnapshotCondition, errType quarkv1alpha1.NetAppVolumeSnapshotConditionType) string {
	for _, cond := range conditions {
		if cond.Type == errType && cond.Status == corev1.ConditionTrue {
			return cond.Message
		}
	}
	return ""
}

// constructs fatal error from error message
func createFatalError(errMsg string) (err *NVCClientFatalError, retryErr bool) {
	if util.VerifyRetryErrorMessage(errMsg) {
		retryErr = true
		errMsg = util.GetRetryErrorMessage(errMsg)
	}
	err = NewNVCClientFatalError(errMsg)
	return
}

func getFatalVolumeOnlineErrors() []string {
	return []string{string(quarkv1alpha1.NetAppVolumeSMBConfigurationError), string(quarkv1alpha1.NetAppVolumeExportPolicyError), string(quarkv1alpha1.NetAppVolumeSnapshotPolicyError), string(quarkv1alpha1.NetAppVolumeCreateError)}
}

func getFatalStoragePoolOnlineErrors() []string {
	return []string{
		string(quarkv1alpha1.NetAppStoragePoolAddressAllocationError),
		string(quarkv1alpha1.NetAppStoragePoolCMEKConfigurationError),
		string(quarkv1alpha1.NetAppStoragePoolAddressExhaustionError),
	}
}
